declare module 'websocket' {
    class w3cwebsocket {
        constructor(url: string);
        onopen(): void;
        onclose: (() => void) | null;
        onmessage(event: {
            data: string;
        }): void;
        send(message: string): void;
        close(): void;
        CONNECTING: string;
        OPEN: string;
        readyState: string;
    }
}
/**
 * Connection to the backend.
 *
 * The standard template is to create a connection first, then use it to
 * wire all UI elements, to add custom callback functions and at last to run
 * [[Connection.connect]] to create a WebSocket connection to the backend
 * server (see example below).
 *
 * The other two essential functions to know about are
 * [[Connection.on]] and [[Connection.emit]].
 *
 * Logging across frontend and backend can be done by emitting `log`,
 * `warn` or `error`, e.g. `emit('log', 'Hello World')`.
 *
 * ```js
 * var databench = new Databench.Connection();
 * Databench.ui.wire(databench);
 *
 *
 * // put custom d.on() methods here
 *
 *
 * databench.connect();
 * ```
 */
export declare class Connection {
    wsUrl: string;
    requestArgs?: string;
    analysisId?: string;
    databenchBackendVersion?: string;
    analysesVersion?: string;
    errorCB: (message?: string) => void;
    private onCallbacks;
    private onProcessCallbacks;
    private preEmitCallbacks;
    private connectCallback;
    private wsReconnectAttempt;
    private wsReconnectDelay;
    private socket?;
    private socketCheckOpen?;
    /**
     * @param  wsUrl        URL of WebSocket endpoint or undefined to guess it.
     * @param  requestArgs  `search` part of request url or undefined to take from
     *                      `window.location.search`.
     * @param  analysisId   Specify an analysis id or undefined to have one generated.
     *                      The connection will try to connect to a previously created
     *                      analysis with that id.
     */
    constructor(wsUrl?: string, requestArgs?: string, analysisId?: string);
    static guessWSUrl(): string;
    /** initialize connection */
    connect(callback?: (connection: Connection) => void): Connection;
    /** close connection */
    disconnect(): void;
    wsCheckOpen(): void;
    wsOnOpen(): void;
    wsOnClose(): void;
    /**
     * Trigger all callbacks for this signal with this message.
     * @param signal  Name of the signal to trigger.
     * @param message Payload for the triggered signal.
     */
    trigger(signal: string, message?: any): void;
    wsOnMessage(event: {
        data: string;
    }): void;
    /**
     * Register a callback that listens for a signal.
     *
     * The signal can be a simple string (the name for a signal/action), but it
     * can also be an Object of the form `{data: 'current_value'}` which would
     * trigger on `data` actions that are sending a JSON dictionary that contains
     * the key `current_value`. In this case, the value that is
     * given to the callback function is the value assigned to `current_value`.
     *
     * ~~~
     * d.on('data', value => { console.log(value); });
     * // If the backend sends an action called 'data' with a message
     * // {current_value: 3.0}, this function would log `{current_value: 3.0}`.
     * ~~~
     *
     * ~~~
     * d.on({data: 'current_value'}, value => { console.log(value); });
     * // If the backend sends an action called 'data' with a
     * // message {current_value: 3.0}, this function would log `3.0`.
     * // This callback is not triggered when the message does not contain a
     * // `current_value` key.
     * ~~~
     *
     * @param  signal    Signal name to listen for.
     * @param  callback  A callback function that takes the attached data.
     */
    on(signal: string | {
        [field: string]: string;
    } | {
        [field: string]: RegExp;
    }, callback: (message: any, key?: string) => void): Connection;
    /**
     * Listen for a signal once.
     *
     * Similar to [on] but returns a `Promise` instead of taking a callback.
     * This is mostly useful for unit tests.
     *
     * @param signal Signal name to listen for.
     */
    once(signal: string | {
        [field: string]: string;
    } | {
        [field: string]: RegExp;
    }): Promise<{
        message: any;
        key?: string;
    }>;
    _on_object(signal: {
        [field: string]: string;
    } | {
        [field: string]: RegExp;
    }, callback: (message: any, key?: string) => void): Connection;
    /**
     * Set a pre-emit hook.
     * @param signalName  A signal name.
     * @param callback    Callback function.
     */
    preEmit(signalName: string, callback: (message: any) => any): Connection;
    /**
     * Emit a signal/action to the backend.
     * @param  signalName  A signal name. Usually an action name.
     * @param  message     Payload attached to the action.
     */
    emit(signalName: string, message?: any): Connection;
    onProcess(processID: number, callback: (message: any) => void): Connection;
}
/**
 * Create a Connection and immediately connect to it.
 *
 * This is a shorthand for
 * ~~~
 * new Connection(wsUrl, requestArgs, analysisId).connect();
 * ~~~
 *
 * Use this function in tests where you know that `connect()` will not trigger
 * any callbacks that you should listen to. In regular code, it is better
 * to define all `on` callbacks before calling `connect()` and so this
 * shorthand should not be used.
 *
 * @param wsUrl       URL of WebSocket endpoint or undefined to guess it.
 * @param requestArgs `search` part of request url or undefined to take from
 *                     `window.location.search`.
 * @param analysisId  Specify an analysis id or undefined to have one generated.
 *                    The connection will try to connect to a previously created
 *                    analysis with that id.
 * @param callback    Called when connection is established.
 */
export declare function connect(wsUrl?: string, requestArgs?: string, analysisId?: string, callback?: (connection: Connection) => void): Connection;
/**
 * Attach to a backend.
 *
 * Similar to [connect](globals.html#connect). Instead of a callback parameter, it
 * returns a Promise that resolves to a [[Connection]] instance once the connection is
 * established.
 *
 * @param wsUrl       URL of WebSocket endpoint or undefined to guess it.
 * @param requestArgs `search` part of request url or undefined to take from
 *                    `window.location.search`.
 * @param analysisId  Specify an analysis id or undefined to have one generated.
 *                    The connection will try to connect to a previously created
 *                    analysis with that id.
 */
export declare function attach(wsUrl?: string, requestArgs?: string, analysisId?: string): Promise<Connection>;
export {};
