"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var websocket_1 = require("websocket");
/**
 * Connection to the backend.
 *
 * The standard template is to create a connection first, then use it to
 * wire all UI elements, to add custom callback functions and at last to run
 * [[Connection.connect]] to create a WebSocket connection to the backend
 * server (see example below).
 *
 * The other two essential functions to know about are
 * [[Connection.on]] and [[Connection.emit]].
 *
 * Logging across frontend and backend can be done by emitting `log`,
 * `warn` or `error`, e.g. `emit('log', 'Hello World')`.
 *
 * ```js
 * var databench = new Databench.Connection();
 * Databench.ui.wire(databench);
 *
 *
 * // put custom d.on() methods here
 *
 *
 * databench.connect();
 * ```
 */
var Connection = /** @class */ (function () {
    /**
     * @param  wsUrl        URL of WebSocket endpoint or undefined to guess it.
     * @param  requestArgs  `search` part of request url or undefined to take from
     *                      `window.location.search`.
     * @param  analysisId   Specify an analysis id or undefined to have one generated.
     *                      The connection will try to connect to a previously created
     *                      analysis with that id.
     */
    function Connection(wsUrl, requestArgs, analysisId) {
        var _this = this;
        this.wsUrl = wsUrl || Connection.guessWSUrl();
        this.requestArgs = (!requestArgs && (typeof window !== 'undefined')) ?
            window.location.search : requestArgs;
        this.analysisId = analysisId;
        this.errorCB = function (msg) { return (msg != null ? console.log("connection error: " + msg) : null); };
        this.onCallbacks = {};
        this.onProcessCallbacks = {};
        this.preEmitCallbacks = {};
        this.connectCallback = function (connection) { };
        this.wsReconnectAttempt = 0;
        this.wsReconnectDelay = 100.0;
        // wire log, warn, error messages into console outputs
        ['log', 'warn', 'error'].forEach(function (wireSignal) {
            _this.on(wireSignal, function (message) { return console[wireSignal]('backend: ', message); });
            _this.preEmit(wireSignal, function (message) {
                console[wireSignal]('frontend: ', message);
                return message;
            });
        });
    }
    Connection.guessWSUrl = function () {
        if (typeof location === 'undefined')
            return '';
        var WSProtocol = location.origin.indexOf('https://') === 0 ? 'wss' : 'ws';
        var path = location.pathname.substring(0, location.pathname.lastIndexOf('/'));
        return WSProtocol + "://" + document.domain + ":" + location.port + path + "/ws";
    };
    /** initialize connection */
    Connection.prototype.connect = function (callback) {
        var _this = this;
        if (!this.wsUrl)
            throw Error('Need a wsUrl.');
        this.connectCallback = callback ? callback : function () { return _this; };
        this.socket = new websocket_1.w3cwebsocket(this.wsUrl);
        this.socketCheckOpen = setInterval(this.wsCheckOpen.bind(this), 2000);
        this.socket.onopen = this.wsOnOpen.bind(this);
        this.socket.onclose = this.wsOnClose.bind(this);
        this.socket.onmessage = this.wsOnMessage.bind(this);
        return this;
    };
    /** close connection */
    Connection.prototype.disconnect = function () {
        if (this.socketCheckOpen) {
            clearInterval(this.socketCheckOpen);
            this.socketCheckOpen = undefined;
        }
        if (this.socket) {
            this.socket.onclose = null;
            this.socket.close();
            this.socket = undefined;
        }
    };
    Connection.prototype.wsCheckOpen = function () {
        if (!this.socket)
            return;
        if (this.socket.readyState === this.socket.CONNECTING) {
            return;
        }
        if (this.socket.readyState !== this.socket.OPEN) {
            this.errorCB('Connection could not be opened. ' +
                'Please <a href="javascript:location.reload(true);" ' +
                'class="alert-link">reload</a> this page to try again.');
        }
        if (this.socketCheckOpen) {
            clearInterval(this.socketCheckOpen);
            this.socketCheckOpen = undefined;
        }
    };
    Connection.prototype.wsOnOpen = function () {
        if (!this.socket)
            return;
        this.wsReconnectAttempt = 0;
        this.wsReconnectDelay = 100.0;
        this.errorCB(); // clear errors
        this.socket.send(JSON.stringify({
            __connect: this.analysisId ? this.analysisId : null,
            __request_args: this.requestArgs,
        }));
    };
    Connection.prototype.wsOnClose = function () {
        if (this.socketCheckOpen) {
            clearInterval(this.socketCheckOpen);
            this.socketCheckOpen = undefined;
        }
        this.wsReconnectAttempt += 1;
        this.wsReconnectDelay *= 2;
        if (this.wsReconnectAttempt > 5) {
            this.errorCB('Connection closed. ' +
                'Please <a href="javascript:location.reload(true);" ' +
                'class="alert-link">reload</a> this page to reconnect.');
            return;
        }
        var actualDelay = 0.7 * this.wsReconnectDelay + 0.3 * Math.random() *
            this.wsReconnectDelay;
        console.log("WebSocket reconnect attempt " + this.wsReconnectAttempt + " " +
            ("in " + actualDelay.toFixed(0) + "ms."));
        setTimeout(this.connect.bind(this), actualDelay);
    };
    /**
     * Trigger all callbacks for this signal with this message.
     * @param signal  Name of the signal to trigger.
     * @param message Payload for the triggered signal.
     */
    Connection.prototype.trigger = function (signal, message) {
        this.onCallbacks[signal].forEach(function (cb) { return cb(message, signal); });
    };
    Connection.prototype.wsOnMessage = function (event) {
        var message = JSON.parse(event.data);
        // connect response
        if (message.signal === '__connect') {
            this.analysisId = message.load.analysis_id;
            this.databenchBackendVersion = message.load.databench_backend_version;
            var newVersion = message.load.analyses_version;
            if (this.analysesVersion && this.analysesVersion !== newVersion) {
                location.reload();
            }
            this.analysesVersion = newVersion;
            this.connectCallback(this);
        }
        // processes
        if (message.signal === '__process') {
            var id = message.load.id;
            var status_1 = message.load.status;
            this.onProcessCallbacks[id].forEach(function (cb) { return cb(status_1); });
        }
        // normal message
        if (message.signal in this.onCallbacks) {
            this.trigger(message.signal, message.load);
        }
    };
    /**
     * Register a callback that listens for a signal.
     *
     * The signal can be a simple string (the name for a signal/action), but it
     * can also be an Object of the form `{data: 'current_value'}` which would
     * trigger on `data` actions that are sending a JSON dictionary that contains
     * the key `current_value`. In this case, the value that is
     * given to the callback function is the value assigned to `current_value`.
     *
     * ~~~
     * d.on('data', value => { console.log(value); });
     * // If the backend sends an action called 'data' with a message
     * // {current_value: 3.0}, this function would log `{current_value: 3.0}`.
     * ~~~
     *
     * ~~~
     * d.on({data: 'current_value'}, value => { console.log(value); });
     * // If the backend sends an action called 'data' with a
     * // message {current_value: 3.0}, this function would log `3.0`.
     * // This callback is not triggered when the message does not contain a
     * // `current_value` key.
     * ~~~
     *
     * @param  signal    Signal name to listen for.
     * @param  callback  A callback function that takes the attached data.
     */
    Connection.prototype.on = function (signal, callback) {
        if (typeof signal === 'object') {
            this._on_object(signal, callback);
            return this;
        }
        if (!(signal in this.onCallbacks))
            this.onCallbacks[signal] = [];
        this.onCallbacks[signal].push(callback);
        return this;
    };
    /**
     * Listen for a signal once.
     *
     * Similar to [on] but returns a `Promise` instead of taking a callback.
     * This is mostly useful for unit tests.
     *
     * @param signal Signal name to listen for.
     */
    Connection.prototype.once = function (signal) {
        var _this = this;
        return new Promise(function (resolve) {
            _this.on(signal, resolve);
        });
    };
    Connection.prototype._on_object = function (signal, callback) {
        var _this = this;
        Object.keys(signal).forEach(function (signalName) {
            var entryName = signal[signalName];
            var filteredCallback = function (data, signalName) {
                Object.keys(data).forEach(function (dataKey) {
                    if (dataKey.match(entryName) === null)
                        return;
                    callback(data[dataKey], dataKey);
                });
            };
            _this.on(signalName, filteredCallback);
        });
        return this;
    };
    /**
     * Set a pre-emit hook.
     * @param signalName  A signal name.
     * @param callback    Callback function.
     */
    Connection.prototype.preEmit = function (signalName, callback) {
        if (!(signalName in this.preEmitCallbacks))
            this.preEmitCallbacks[signalName] = [];
        this.preEmitCallbacks[signalName].push(callback);
        return this;
    };
    /**
     * Emit a signal/action to the backend.
     * @param  signalName  A signal name. Usually an action name.
     * @param  message     Payload attached to the action.
     */
    Connection.prototype.emit = function (signalName, message) {
        var _this = this;
        // execute preEmit hooks before sending message to backend
        if (signalName in this.preEmitCallbacks) {
            this.preEmitCallbacks[signalName].forEach(function (cb) {
                message = cb(message);
            });
        }
        // socket will never be open
        if (!this.socket)
            return this;
        // socket is not open yet
        if (this.socket.readyState !== this.socket.OPEN) {
            setTimeout(function () { return _this.emit(signalName, message); }, 5);
            return this;
        }
        // send to backend
        this.socket.send(JSON.stringify({ signal: signalName, load: message }));
        return this;
    };
    Connection.prototype.onProcess = function (processID, callback) {
        if (!(processID in this.onProcessCallbacks)) {
            this.onProcessCallbacks[processID] = [];
        }
        this.onProcessCallbacks[processID].push(callback);
        return this;
    };
    return Connection;
}());
exports.Connection = Connection;
/**
 * Create a Connection and immediately connect to it.
 *
 * This is a shorthand for
 * ~~~
 * new Connection(wsUrl, requestArgs, analysisId).connect();
 * ~~~
 *
 * Use this function in tests where you know that `connect()` will not trigger
 * any callbacks that you should listen to. In regular code, it is better
 * to define all `on` callbacks before calling `connect()` and so this
 * shorthand should not be used.
 *
 * @param wsUrl       URL of WebSocket endpoint or undefined to guess it.
 * @param requestArgs `search` part of request url or undefined to take from
 *                     `window.location.search`.
 * @param analysisId  Specify an analysis id or undefined to have one generated.
 *                    The connection will try to connect to a previously created
 *                    analysis with that id.
 * @param callback    Called when connection is established.
 */
function connect(wsUrl, requestArgs, analysisId, callback) {
    return new Connection(wsUrl, requestArgs, analysisId).connect(callback);
}
exports.connect = connect;
/**
 * Attach to a backend.
 *
 * Similar to [connect](globals.html#connect). Instead of a callback parameter, it
 * returns a Promise that resolves to a [[Connection]] instance once the connection is
 * established.
 *
 * @param wsUrl       URL of WebSocket endpoint or undefined to guess it.
 * @param requestArgs `search` part of request url or undefined to take from
 *                    `window.location.search`.
 * @param analysisId  Specify an analysis id or undefined to have one generated.
 *                    The connection will try to connect to a previously created
 *                    analysis with that id.
 */
function attach(wsUrl, requestArgs, analysisId) {
    var connection = new Connection(wsUrl, requestArgs, analysisId);
    return new Promise(function (resolve) { return connection.connect(resolve); });
}
exports.attach = attach;
//# sourceMappingURL=connection.js.map