/**
 * This is a basic set of UI elements to create analyses without having to add
 * frontend frameworks like Angular or React.
 *
 * @module ui
 */
import { Connection } from './connection';
export declare namespace ui {
    interface HTMLDatabenchElement extends HTMLElement {
        databenchUI: UIElement;
    }
    /**
     * Abstract class for user interface elements which provides general helpers
     * to determine the action name from an HTML tag and a way to modify the message
     * that is sent with actions of wired elements.
     *
     * The constructor adds the variables `actionName` and `wireSignal` using
     * [[UIElement.determineActionName]] and
     * [[UIElement.determineWireSignal]] respectively.
     * It also adds `this` UI element to the DOM node at `databenchUI`.
     */
    class UIElement {
        node: HTMLDatabenchElement;
        static idCounter: number;
        idCount: number;
        actionName: string;
        wireSignal: string | {
            [x: string]: string;
        };
        /**
         * @param node  An HTML element.
         */
        constructor(node: HTMLElement | HTMLDatabenchElement);
        /**
         * Formats the payload of an action.
         * @param  value Original payload.
         * @return       Modified payload.
         */
        actionFormat(value: any): any;
        /**
         * Determine whether to skip this element.
         *
         * This can be forced by adding a `data-skipwire=true` attribute
         * to the HTML tag.
         *
         * @param node A HTML element.
         */
        static skipWire(node: HTMLElement): boolean;
        /**
         * Determine the name of the action that should be associated with the node.
         *
         * The action name is determined from
         * the tag's `data-action`, `name` or `id` attribute (in that order).
         *
         * @param  node A HTML element.
         * @return      Name of action or null.
         */
        determineActionName(node: HTMLElement): string;
        /**
         * Determine the name of the signal that should be listened to from the backend.
         *
         * The signal name is determined from the `data-signal`,
         * `data-action`, `name` or `id` attribute (in that order.)
         * For all attributes apart from `data-signal`, the value is wrapped in an
         * object like `{ data: value-of-attribute }`. The `data-signal` value
         * can contain a `:` which will be used to create an object as well. That means
         * that `data-signal="data:myvalue"` gives the same result as `data-action="myvalue"`.
         *
         * @param  node A HTML element.
         * @return      Name of a signal or null.
         */
        determineWireSignal(node: HTMLElement): string | {
            [x: string]: string;
        };
    }
    /**
     * Shows all `console.log()` messages and `log` actions from backend.
     *
     * Usually wired to a `<pre id="log">` element.
     */
    class Log extends UIElement {
        limitNumber: number;
        limitLength: number;
        _messages: string[][];
        /**
         * @param  node            Primary node.
         * @param  limitNumber     Maximum number of messages to show.
         * @param  limitLength     Maximum length of a message.
         */
        constructor(node: HTMLElement, limitNumber?: number, limitLength?: number);
        render(): Log;
        add(message: any, source?: string): Log;
        /** Wire all logs. */
        static wire(conn: Connection, id?: string, wireSignals?: string[], limitNumber?: number, limitLength?: number): void;
    }
    /**
     * Visual representation of alerts like connection failures.
     *
     * Usually wired to a `<div id="databench-alerts">` element.
     */
    class StatusLog extends UIElement {
        formatter: (message: string, count: number) => string;
        _messages: {
            [message: string]: number;
        };
        /**
         * @param  node       HTML node.
         * @param  formatter  Formats a message and a count to a string.
         */
        constructor(node: HTMLElement, formatter?: (message: string, count: number) => string);
        /**
         * The default formatter function
         * @param  message   A message.
         * @param  count     Count of the message.
         * @returns HTML formatted version of the inputs.
         */
        static defaultAlert(message: string, count: number): string;
        render(): StatusLog;
        add(message: any): StatusLog;
        /** Wire all status logs. */
        static wire(conn: Connection, id?: string, formatter?: (message: string, count: number) => string): void;
    }
    enum ButtonState {
        Idle = 1,
        Active = 2,
    }
    /**
     * A button, and usually wired to any `<button>` with an action name.
     *
     * This button also binds to process IDs of the backend. That means
     * that the button is disabled (using the CSS class `disabled`) while the
     * backend is processing the action that got started when it was clicked.
     * A simple example is below.
     *
     * In `index.html`:
     * ```html
     * <button data-action="run">Run</button>
     * ```
     *
     * In `analysis.py`:
     * ```py
     * @ databench.on
     * def run(self):
     *     """Run when button is pressed."""
     *     pass
     * ```
     *
     * In this form, Databench finds the button automatically and connects it
     * to the backend. No additional JavaScript code is required.
     */
    class Button extends UIElement {
        _state: ButtonState;
        /**
         * @param node   DOM node to connect.
         */
        constructor(node: HTMLElement);
        /**
         * Called on click events. When a button is wired, this function is overwritten
         * with the actual function that is triggered on click events.
         * @param  processID   a random id for the process that could be started
         */
        clickCB(processID: number): void;
        render(): Button;
        click(): Button;
        state(s: ButtonState): Button;
        /** Wire all buttons. */
        static wire(conn: Connection, root?: Document | HTMLElement): void;
    }
    /**
     * Data bound text elements.
     *
     * Wired to `<span>`, `<p>`, `<div>`, `<i>` and `<b>` tags with a
     * `data-action` attribute specifying the action name.
     */
    class Text extends UIElement {
        /**
         * Format the value.
         *
         * The default implementation returns a `JSON.stringify()` version for values
         * that are `typeof value == 'object'`. Otherwise it passes the value through.
         *
         * @param  value  Value as represented in the backend.
         * @return Formatted representation of the value.
         */
        formatFn(value: any): string;
        /** Reads the value. */
        getValue(): string;
        /** Reads the value. */
        setValue(v?: string): Text;
        /**
         * Wire all text.
         * @param  {Connection} conn Connection to use.
         */
        static wire(conn: Connection, root?: Document | HTMLElement): void;
    }
    /** Make an `<input[type='text']>` with an action name interactive. */
    class TextInput extends UIElement {
        node: HTMLInputElement & HTMLDatabenchElement;
        _triggerOnKeyUp: boolean;
        /**
         * @param node  The node to connect.
         */
        constructor(node: HTMLElement);
        /**
         * Format the value.
         * @param  value  Value as represented in the backend.
         * @return Formatted representation of the value.
         */
        formatFn(value: any): string;
        /**
         * Callback that is triggered on frontend changes.
         * @param  value  A formatted action.
         */
        changeCB(value: any): void;
        change(): void;
        /**
         * The default is `false`, which means that the callback is only triggered on
         * `change` events (i.e. pressing enter or unfocusing the element).
         * Setting this to true will trigger on every `keyup` event of this element.
         *
         * @param  v Whether to trigger on `keyup` events. Default is true.
         * @return self
         */
        triggerOnKeyUp(v: boolean): TextInput;
        /** Reads and sets the value. */
        getValue(): string;
        /** Set value */
        setValue(v: string): TextInput;
        /** Wire all text inputs. */
        static wire(conn: Connection, root?: Document | HTMLElement): void;
    }
    /**
     * Make all `<input[type='range']>` with an action name interactive.
     *
     * In `index.html`:
     * ```html
     * <label for="samples">Samples:</label>
     * <input type="range" id="samples" value="1000" min="100" max="10000" step="100" />
     * ```
     *
     * In `analysis.py`:
     * ```py
     * @ databench.on
     * def samples(self, value):
     *     """Sets the number of samples to generate per run."""
     *     yield self.set_state(samples=value)
     * ```
     */
    class Slider extends UIElement {
        node: HTMLInputElement & HTMLDatabenchElement;
        labelNode: HTMLElement | undefined;
        labelHtml: string;
        /**
         * @param  node       DOM node to bind.
         * @param  labelNode  DOM node label that corresponds to the slider.
         */
        constructor(node: HTMLElement, labelNode?: HTMLElement);
        /**
         * Callback with changes to the slider value.
         * @param  value   Value from a sliderToValue() transform.
         */
        changeCB(value: number): void;
        /**
         * Transform a backend value to a slider value.
         * @param   value  Value as stored in backend.
         * @return  Value for the HTML range element.
         */
        valueToSlider(value: number): string;
        /**
         * Transform a value from the HTML range element to a value that should be stored.
         * @param  s  Value from HTML range element.
         * @return Value to store.
         */
        sliderToValue(s: number): number;
        /**
         * How a value should be represented.
         * For example, this can add units or convert from radians to degrees.
         * @param  value  Input value as it is stored in the backend.
         * @return Representation of a value.
         */
        formatFn(value: number): string;
        render(): Slider;
        /** Reads the value. */
        getValue(): number;
        /** Set the value. */
        setValue(v: number): Slider;
        change(): void;
        /** Find all labels for slider elements. */
        static labelsForSliders(root: Document | HTMLElement): {
            [inputname: string]: HTMLLabelElement;
        };
        /** Wire all sliders. */
        static wire(conn: Connection, root?: Document | HTMLElement): void;
    }
    /**
     * Connect an `<img>` with a signal name to the backend.
     *
     * The signal message is placed directly into the `src` attribute of the image
     * tag. For matplotlib, that formatting can be done with the utility function
     * `fig_to_src()` (see example below).
     *
     * In `index.html`:
     * ```html
     * <img alt="my plot" data-signal="mpl" />
     * ```
     *
     * In `analysis.py`:
     * ```py
     * import matplotlib.pyplot as plt
     * ...
     * fig = plt.figure()
     * ...
     * self.emit('mpl', databench.fig_to_src(fig))
     * ```
     */
    class Image extends UIElement {
        node: HTMLImageElement & HTMLDatabenchElement;
        getSrc(): string;
        /** Sets theSrc. */
        setSrc(v: string): Image;
        /** Wire all text inputs. */
        static wire(conn: Connection, root?: Document | HTMLElement): void;
    }
    /**
     * Wire all the UI elements to the backend. The action name is determined by
     * [[UIElement.determineActionName]] and the action message can be modified
     * by overwriting [[UIElement.actionFormat]]. The signal name is determined by
     * [[UIElement.determineWireSignal]].
     *
     * @param  connection  A Databench.Connection instance.
     * @return The same connection.
     */
    function wire(connection: Connection, root?: Document | HTMLElement): Connection;
}
